@@ -18,8 +18,6 @@
 # Date:		Tue Mar 17 13:44:00 1998
 # Description:  Perl traceroute module for performing traceroute(1)
 #		functionality.
-#
-# $Id: Traceroute.pm,v 1.22 2005/07/09 20:22:58 hag Exp $
 
 # Currently attempts to parse the output of the system traceroute command,
 # which it expects will behave like the standard LBL traceroute program.
@@ -46,9 +44,11 @@ use IO::Pipe;
 use IO::Select;
 use Socket;
 use Symbol qw(qualify_to_ref);
+use Time::HiRes qw(time);
+use Errno qw(EAGAIN EINTR);
 use Data::Dumper;		# Debugging
 
-$VERSION = "1.09";		# Version number is only incremented by
+$VERSION = "1.11";		# Version number is only incremented by
 				# hand.
 
 @ISA = qw(Exporter);
@@ -100,6 +100,7 @@ my @public_instance_vars =
        queries
        query_timeout
        source_address
+       text
        trace_program
        timeout
        no_fragment
@@ -124,9 +125,6 @@ use constant query_time_offset => 2;
 
 # Constructor
 
-# XXX Stanley Hopcroft is having issues with timeouts.  Track down his
-# email and produce a test case.
-
 sub new ($;%) {
     my $self = shift;
     my $type = ref($self) || $self;
@@ -185,6 +183,8 @@ sub init {
 
     if(defined($self->host)) {
 	$self->traceroute;
+    } elsif(defined($self->text)) {
+	$self->_parse($self->text)
     }
 
     $self->debug_print(9, Dumper($self));
@@ -222,6 +222,8 @@ sub clone ($;%) {
 
     if(defined($clone->host)) {
 	$clone->traceroute;
+    } elsif(defined($clone->text)) {
+	$clone->_parse($clone->text)
     }
 
     $clone->debug_print(9, Dumper($clone));
@@ -246,10 +248,16 @@ sub traceroute ($) {
     # waiting for a timeout if we need to.  Accumulate the text for
     # parsing later in one fell swoop.
 
-    # Note time
-    my $start_time = time();
+    # Note time.  Time::HiRes will give us floating point.
+    my $start_time;
+    my $end_time;
     my $total_wait = $self->timeout();
-    my @this_wait = defined($total_wait) ? ($total_wait) : ();
+    my @this_wait;
+    if(defined($total_wait)) {
+	$start_time = time();
+	push(@this_wait, $total_wait);
+	$end_time = $start_time + $total_wait;
+    }
 
     my $tr_pipe = $self->_make_pipe();
     my $select = new IO::Select($tr_pipe);
@@ -263,28 +271,47 @@ sub traceroute ($) {
 	my $fh;
 	foreach $fh (@ready) {
 	    my $buf;
-	    my $len = $fh->read($buf, 32);
-	    die "read error: $!" unless(defined($len));
+	    my $len = $fh->sysread($buf, 2048);
+
+	    # XXX Linux is fond of returning EAGAIN, which we'll need
+	    # to check for here.  Still true for sysread?
+	    if(!defined($len)) {
+		my $errno = int($!);
+		next out if(($errno == EAGAIN) || ($errno == EINTR));
+		die "read error: $!";
+	    }
 	    last out if(!$len);	# EOF
-	    $self->_text_accumulator($buf);
+
+	    $self->text($self->text() . $buf);
 	}
 
-	# Check for timeout
-	my $now = time();
+	# Adjust select timer if we need to.
 	if(defined($total_wait)) {
-	    if($now > ($start_time + $total_wait)) {
-		$self->stat(TRACEROUTE_TIMEOUT);
-		last out;
-	    }
-	    $this_wait[0] = ($start_time + $total_wait) - $now;
+	    my $now = time();
+	    last out if($now >= $end_time);
+	    $this_wait[0] = $end_time - $now;
 	}
     }
+    if(defined($total_wait)) {
+	my $now = time();
+	$self->stat(TRACEROUTE_TIMEOUT)	if($now >= $end_time);
+
+	# This is exceedingly dubious.  Crawl into IO::Pipe::End's
+	# innards, and nuke the pid connected to our pipe.  Otherwise,
+	# close will call waitpid, which we certainly don't wait for a
+	# timeout.
+	delete ${*$tr_pipe}{io_pipe_pid};
+    }
 
     $tr_pipe->close();
 
-    # Do the grunt parsing work
-    $self->_parse($self->_text_accumulator());
+    my $accum = $self->text();
+    die "No output from traceroute.  Exec failure?" if($accum eq "");
 
+    # Do the grunt parsing work
+    $self->_parse($accum);
+
+    # XXX are you really sure you want to do it like this??
     if($self->stat() != TRACEROUTE_TIMEOUT) {
 	$self->stat(TRACEROUTE_OK);
     }
@@ -292,6 +319,12 @@ sub traceroute ($) {
     $self;
 }
 
+sub parse {
+    my $self = shift;
+
+    $self->_parse($self->text());
+}
+
 ##
 # Hop and query functions
 
@@ -411,11 +444,8 @@ sub _make_pipe ($) {
     open(STDERR, ">& SAVESTDERR");
     close(SAVESTDERR);
 
-    # XXX We're going to assume that an eof right after fork/exec is
-    # actually a failure.  This is quite dubious.
-    if($result->eof) {
-  	die "No output from traceroute.  Exec failure?";
-    }
+    # Long standing bug; the pipe needs to be marked non blocking.
+    $result->blocking(0);
 
     $result;
 }
@@ -476,23 +506,12 @@ sub _parse ($$) {
     my $self = shift;
     my $tr_output = shift;
 
-    my @list_of_lines = split(/\n/, $tr_output);
-    my $counter = 0;
-    my $max_limit = @list_of_lines; 
-    # find whether timeout is defined 
-    my $total_wait = $self->timeout();
-    my @this_wait = defined($total_wait) ? ($total_wait) : ();
-
     # This is a crufty hand coded parser that does its job well
     # enough.  The approach of regular expressions without state is
     # far from perfect, but it gets the job done.
   line:
     foreach $_ (split(/\n/, $tr_output)) {
-	# dont parse the last line because timeout might have introduced garbage into it 
-	$counter ++;
-	if ($counter == $max_limit  && $total_wait) {
-		last;
-	}
+
 	# Some traceroutes appear to print informational line to stdout,
 	# and we don't care.
 	/^traceroute to / && next;
@@ -539,12 +558,17 @@ sub _parse ($$) {
 		$_ = substr($_, length($&));
 		next query;
 	    };
-	    # ipv6 address of a response
-	    /^ ([0-9a-fA-F:]+)/ && do {
+	    # ipv6 address of a response.  This regexp is sleazy.
+	    /^ ([0-9a-fA-F:]*:[0-9a-fA-F]*(?:\.\d+\.\d+\.\d+)?)/ && do {
 		$addr = $1;
 		$_ = substr($_, length($&));
 		next query;
 	    };
+	    # Redhat FC5 traceroute does this; it's redundant.
+	    /^ \((\d+\.\d+\.\d+\.\d+)\)/ && do {
+		$_ = substr($_, length($&));
+		next query;
+	    };
 	    # round trip time of query
 	    /^   ?([0-9.]+) ms/ && do {
 		$time = $1 + 0;
@@ -623,20 +647,11 @@ sub _parse ($$) {
     }
 }
 
-# I don't understand why this one won't work with the accessor generator.
-sub _text_accumulator ($;$) {
-    my $self = shift;
-    my $name = "_text_accumulator";
-    my $old = $self->{$name};
-    $self->{$name} = $self->{$name}.$_[0] if @_;
-    return $old;
-}
-
 sub _zero_text_accumulator ($) {
     my $self = shift;
-    my $elem = "_text_accumulator";
+    my $elem = "text";
 
-    delete $self->{$elem};
+    $self->{$elem} = "";
 }
 
 # Hop stuff
@@ -730,7 +745,7 @@ Net::Traceroute - traceroute(1) functionality in perl
 =head1 SYNOPSIS
 
     use Net::Traceroute;
-    $tr = Net::Traceroute->new(host=> "life.ai.mit.edu");
+    $tr = Net::Traceroute->new(host => "life.ai.mit.edu");
     if($tr->found) {
 	my $hops = $tr->hops;
 	if($hops > 1) {
@@ -744,14 +759,17 @@ Net::Traceroute - traceroute(1) functionality in perl
 This module implements traceroute(1) functionality for perl5.  It
 allows you to trace the path IP packets take to a destination.  It is
 currently implemented as a parser around the system traceroute
-command.
+command.  It has two basic modes of operation, one, where it will run
+traceroute for you, and the other where you provide text from
+previously runing traceroute to parse.
 
 =head1 OVERVIEW
 
 A new Net::Traceroute object must be created with the I<new> method.
 Depending on exactly how the constructor is invoked, it may perform
-the traceroute immediately, or it may return a "template" object that
-can be used to set parameters for several subsequent traceroutes.
+some tracing and/or parsing actions immediately, or it may return a
+"template" object that can be used to set parameters for several
+subsequent traceroutes.
 
 Methods are available for accessing information about a given
 traceroute attempt.  There are also methods that view/modify the
@@ -767,6 +785,7 @@ TIME_EXCEEDED messages.
 				[debug		=> $debuglvl,]
 				[max_ttl	=> $max_ttl,]
 				[host		=> $host,]
+				[text		=> $text,]
 				[queries	=> $queries,]
 				[query_timeout	=> $query_timeout,]
 				[timeout	=> $timeout,]
@@ -778,13 +797,14 @@ TIME_EXCEEDED messages.
     $frob = $obj->clone([options]);
 
 This is the constructor for a new Net::Traceroute object.  If given
-C<host>, it will actually perform the traceroute.  You can call the
-traceroute method later.
+C<host>, it will immediately perform the traceroute.  If given C<text>,
+it will parse that text as traceroute output.
 
 Given an existing Net::Traceroute object $obj as a template, you can
 call $obj->clone() with the usual constructor parameters.  The same
 rules apply about defining host; that is, traceroute will be run if it
-is defined.  You can always pass host => undef to clone.
+is defined, or text will be parsed.  You can always pass
+C<host => undef, text => undef> to clone.
 
 Possible options are:
 
@@ -793,6 +813,9 @@ Traceroute object with no traceroute data in it.  The module always
 uses IP addresses internally and will attempt to lookup host names via
 inet_aton.
 
+B<text> - Output from a previously run traceroute.  If set, and host
+isn't, the given text will be parsed.
+
 B<base_port> - Base port number to use for the UDP queries.
 Traceroute assumes that nothing is listening to port C<base_port> to
 C<base_port + (nhops - 1)>
@@ -844,6 +867,11 @@ packets, rather than UDP.
 Run system traceroute, and parse the results.  Will fill in the rest
 of the object for informational queries.
 
+=item parse
+
+Parse the previously provided C<text>, filling in the rest of the
+object for queries.
+
 =back
 
 =head2 Controlling traceroute invocation
@@ -871,6 +899,8 @@ aren't documented here.
 
 =item host([HOST])
 
+=item text([TEXT])
+
 =item timeout([TIMEOUT])
 
 =item source_address([SRC])
